from random import random, randint, seed 
from ppm import PPM, Color # eigenes PPM-Bildmodul

class Nasch2:

    def __init__(self, p=0.35, d=0.2, vmax=5, roadlen=800, timesteps=600):
        self.roadlen = roadlen   # Anzahl Zellen
        self.timesteps = timesteps # Anzahl Zeitschritte
        self.p = p # Trödelwahrscheinlichkeit
        self.d = d # Fahrzeugdichte
        self.vmin = 0 # Minimaltempo 
        self.vmax = vmax # Höchstgeschwindigkeit (Zellen/Sekunde)
        self.init_min = 0 # Initialgeschwindigkeiten
        self.init_max = 0 # 
        self.road = []   # Fahrbahn: (leere) Liste aus Zellen 
        self.empty = '.' # Symbol für leere Zelle
        self.generate_road()
        self.img = PPM(self.roadlen, self.timesteps)
        self.insert(0)

    # Hilfsfunktion zum Erzeugen der Fahrbahn
    def generate_road(self):
        for i in range(0, self.roadlen):
            if random() < self.d:
                v = self.road.append(randint(self.init_min, self.init_max))
            else:
                self.road.append(self.empty)

    # Überschreibe Textdarstellung der Klasse  
    def __str__(self):
        '''Zeigt maximal die ersten n=50 Zellen der Fahrbahn an'''
        n = 50
        k = min(n, self.roadlen)
        txt = '|' + ''.join(str(c) for c in self.road[:k])
        if self.roadlen <= n:
            txt += '|'
        else:
            txt += '  ...'
        return txt
            
    # Beschleunigen
    def accelerate(self):
        for i in range(0, self.roadlen):
            if self.road[i] != self.empty:
                v = self.road[i]
                if v < self.vmax:
                    self.road[i] = v+1
                
    # Abbremsen
    def slowdown(self):
        for i in range(0, self.roadlen):
            if self.road[i] != self.empty:
                v = self.road[i]
                for j in range(0, v+1):
                    if self.road[(i+1+j)%self.roadlen] != self.empty:
                        break # hat in meiner Lösung gefehlt
                self.road[i] = j  # falsch eingerückt
                
    # Trödeln
    def randomize(self):
        for i in range(0, self.roadlen):
            if self.road[i] != self.empty:
                v = self.road[i]
                if random() < self.p and v > 0:
                    self.road[i] = v-1

    # Bewegen
    def move(self):
        # Hinweise:
        # (1) Zuerst eine neue leere Fahrbahn erzeugen
        # (2) Bewegte Fahrzeuge in diese Fahrbahn schreiben
        # (3) Fahrbahn durch neue Fahrbahn ersetzen
        new_road = [self.empty for cell in self.road]
        for i in range(0, self.roadlen):
            if self.road[i] != self.empty:
                v = self.road[i]
                new_road[(i+v)%self.roadlen] = v
        self.road = new_road

    # Fahrbahnzustand zum Zeitpunkt t ins Bild einfügen
    # Fahrzeug=1 (Schwarz), kein Fahrzeug=0 (Weiss)
    def insert(self, t):
        '''Aktuelle Fahrbahn in die Grafikzeile t einfügen'''
        palette = Color.heat_map_colors(self.vmax)
        for i in range(0, self.roadlen):
            if self.road[i] == self.empty:
                self.img.set_pixel(t, i, Color.White)
            else:
                self.img.set_pixel(t, i, palette[self.road[i]])

    def simulate(self, filename=''):
        '''Führt die Simulation durch'''
        if filename == '':
            w = self.roadlen
            t = self.timesteps
            p = self.p
            v = self.vmax
            d = self.d
            filename = f'nasch-w{w}-t{t}-p{p}-vmax{v}-d{d}.ppm'
        
        for t in range(1, self.timesteps):
            self.accelerate()
            self.slowdown()
            self.randomize()
            self.move()
            self.insert(t)
        # Bildmatrix in PPM-Datei schreiben            
        self.img.write(filename)

############################################################
# Simulationen:
############################################################

seed(4) # Pseudozufallszahlengenerator mit seed=4 starten

r = Nasch2(p=0.35, d=0.2, vmax=5, roadlen=800, timesteps=400)
r.simulate()

